import unittest
import os
import signal
import shutil
from subprocess import Popen

if __name__ == '__main__':
    import sys
    from os.path import dirname, abspath, join
    # add CastorScript/Script to path so imports keeps working
    SCRIPT_DIR = abspath(join(dirname(__file__), '..'))
    sys.path.append(SCRIPT_DIR)

from UnitTests.testUtils import DirectoryChecker, findCastorScriptPids

# I'm a developer, I want to see if the script can copy files.
class TestCanCopyFilesLocally(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        self.timeout = False
        # kill all processes of CastorScript
        # kill the processes based on PID
        for pid in findCastorScriptPids():
            os.kill(int(pid), signal.SIGTERM)

        # Define folder structure
        self.thisdir = os.path.dirname(__file__)

        self.tmp = "tmp"

        self.src = self.tmp + "/src"
        self.des = self.tmp + "/des"
        self.log = self.tmp + "/log"

        self.tmpPath = os.path.abspath(os.path.join(self.thisdir, self.tmp))

        self.srcPath = os.path.abspath(os.path.join(self.thisdir, self.src))
        self.desPath = os.path.abspath(os.path.join(self.thisdir, self.des))
        self.logPath = os.path.abspath(os.path.join(self.thisdir, self.log))

        # Setup the testcase tools
        self.logDirUtil = DirectoryChecker(self.logPath)
        self.srcDirUtil = DirectoryChecker(self.srcPath)
        self.desDirUtil = DirectoryChecker(self.desPath)

        # Setup the timers
        self.timeToTimeout = 5
        self.timerUtil = None

        # I create some folders
        try:
            os.makedirs(self.srcPath)
            os.makedirs(self.desPath)
            os.makedirs(self.logPath)
        except:
            shutil.rmtree(self.tmpPath)
            os.makedirs(self.srcPath)
            os.makedirs(self.desPath)
            os.makedirs(self.logPath)

        # And I populate only the src folder with files to copy
        for x in range(1):
            open(self.srcPath + '/file' + str(x) + '.data', "w").close()
            #open(self.desPath + '/file' + str(x) + '.data', "w").close()
            #open(self.logPath + '/file' + str(x) + '.data', "w").close()

        # I make a config file from the template (CastorScript/Configs)
        # I find the config file
        self.confTemplatePath = os.path.abspath(os.path.join(self.thisdir, "../../Configs"))

        # I open it
        cfgTemplate = open(self.confTemplatePath + "/template.cfg", "r")

        # I make a new file in CastorScript/tmp directory
        cfgTest = open(self.tmpPath + "/" + "funcTestConf.cfg", "w")

        # And I copy the content and replaces the directories with my own dirs
        for line in cfgTemplate:

            if line.find("LogDir: ") > -1:
                cfgTest.write("LogDir: \'" + self.logPath + "\'\n")
                #print(line)
            elif line.find("DirList: ") > -1:
                cfgTest.write("DirList: [\'" + self.srcPath + "\',]\n")
                #print(line)
            elif line.find("CopyDir: ") > -1:
                cfgTest.write("CopyDir: \'" + self.desPath + "\'\n")
                #print(line)
            else:
                cfgTest.write(line)
                #print(line)

        cfgTest.close()

        # Run the script with config
        #NON SHELL

        #self.castorscriptprocess = Popen(scriptcommand)
        scriptcommand = ["python", self.thisdir + "/../CastorScript.py", self.tmpPath + "/funcTestConf.cfg" ]
        self.castorscriptprocess = Popen(scriptcommand)
        #Run in shell
        #self.castorscriptprocess = Popen("python " + self.thisdir + "/../CastorScript.py " + self.tmpPath + "/funcTestConf.cfg", shell=True)
        # To kill the process for debugging
        #self.castorscriptprocess.kill()

    @classmethod
    def tearDownClass(self):
        # stop process
        self.castorscriptprocess.send_signal(signal.SIGUSR2)
        self.castorscriptprocess.wait()

        # remove directories and files
        shutil.rmtree(self.tmpPath)


    # Test cases:

    def test_A_cs_should_be_instanciated(self):
        #test if the script runs by Instance
        self.assertNotEqual(self.castorscriptprocess, None, msg="Castorscript, is not running; Subprocess is of type none")

    def test_B_cs_should_be_running_and_have_a_pid(self):
        #test if the script runs by PID
        self.assertIsNot(len(findCastorScriptPids()), 0, msg="No CastorScript PIDs found, so must not be running")

    def test_C_cs_should_produce_logfiles(self):
        self.timeToTimeout = 10
        self.logDirUtil.waitForAnyFile(self.timeToTimeout)
        self.timeout = self.logDirUtil.timerUtil.timedout
        output = self.logDirUtil.lsOutput
        timeoutMessage = "Timeout, found no log files within a timeframe of " + str(self.timeToTimeout) + " seconds"
        self.assertFalse(self.timeout, msg=timeoutMessage)

        outNotFoundMessage = "No files ending with .out found in log directory, found:" + output
        self.assertTrue(output.find(".out") > -1, msg=outNotFoundMessage)

    def test_D_cs_should_copy_files_to_des_directory(self):
        self.timeToTimeout = 4
        self.desDirUtil.waitForAnyFile(self.timeToTimeout)
        self.timeout = self.desDirUtil.timerUtil.timedout
        output = self.desDirUtil.lsOutput
        timeoutMessage = "Timeout, found no files within a timeframe of " + str(self.timeToTimeout) + " seconds"
        self.assertFalse(self.timeout, msg=timeoutMessage)

        dataNotFoundMessage = "No files ending with .data found in des directory, found:" + output
        self.assertTrue(output.find(".data") > -1, msg=dataNotFoundMessage)

    def test_E_src_files_should_be_deleted(self):
        self.timeToTimeout = 5
        self.srcDirUtil.waitForEmptyDir(self.timeToTimeout)

        self.timeout = self.srcDirUtil.timerUtil.timedout
        output = self.srcDirUtil.lsOutput
        timeoutMessage = "Timeout, found no empty dir within a timeframe of " + str(self.timeToTimeout) + " seconds"

        self.assertFalse(self.timeout, msg=timeoutMessage)
        self.assertTrue(output.find("file") == -1, msg="No files ending with .out found in log directory, found:" + output)


if __name__ == '__main__':
    unittest.main()
