#!/bin/env python

""" @package ManagerThread
This module describe the Manager Thread.
It is in charge of ls on SFO disk.
"""

import threading, os, glob
from time import time
import logging
import datetime
import pprint
import math
import cs.Tools.Constants as Constants
import errno
import re
from cs.Tools.utils import set_log_level,formatter,thread_id_string
from itertools import chain

class ManagerThread(threading.Thread):

    def __init__(self,conf,event,CopyQueue,DeleteQueue,ClearQueue):

        threading.Thread.__init__(self, name="ManagerThread")

        self.parser = ''
        self.conf = conf
        self.event = event
        self.CopyQueue = CopyQueue
        self.DeleteQueue = DeleteQueue
        self.ClearQueue = ClearQueue

        # Get configuration parameters
        self.lockFile = self.conf.lockFile
        self.DirList = [os.path.normpath(k) for k in self.conf.DirList]
        self.DataFilePattern = self.conf.DataFilePattern
        self.nDel = self.conf.nDel
        self.ManagerTimeout = self.conf.ManagerTimeout
        self.MinSizekB = self.conf.MinSizekB
        self.RemoveSmallFiles = self.conf.RemoveSmallFiles
        self.ProblDelay = self.conf.ProblDelay
        self.ProblScalingFactor = self.conf.ProblScalingFactor
        self.ProblDelayLimit = self.conf.ProblDelayLimit
        self.CopyDir = self.conf.CopyDir
        self.Pool = self.conf.Pool
        self.StageHost = self.conf.StageHost
        self.drivenPool = self.conf.drivenPool
        self.LogDir = self.conf.LogDir
        self.LogLevel = self.conf.LogLevel
        self.ERSLogLevel = self.conf.ERSLogLevel
        self.ERSenabled = self.conf.ERSenabled

        self.CopyFileList = [] # all files currently handled by CastorScript
        self.ProblDict = {} # files declared problematic with extra info to manage retries
        self.exitFlag = False
        self.Year = ''
        self.Month = ''

        # Set Logger for ManagerThread
        self.logFile = os.path.join(self.LogDir,'ManagerLog.out')
        self.ManagerLog = logging.FileHandler(self.logFile,'w')
        self.ManagerLog.setFormatter(formatter)
        set_log_level(self.LogLevel, self.ManagerLog)

        self.logger = logging.getLogger('ManagerThread')
        self.logger.addHandler(self.ManagerLog)

        if self.ERSenabled:
            try:
                import ers
                #ers.addLoggingHandler('ManagerThread')
                self.manage_ERS_handler=ers.LoggingHandler()
                set_log_level(self.ERSLogLevel,self.manage_ERS_handler)
                logging.getLogger('ManagerThread').addHandler( self.manage_ERS_handler )
            except Exception,ex:
                self.logger.error("ManagerThread could not create ERS, reason is '%s'", str(ex))


    def initialize(self):
        # Check current year and month for directory streaming
        self.updateDates()

        # Check the validity of all the helper files from previous runs
        self.logger.debug('Check helper files from previous run')

        for folder in self.DirList:
            for oldfile in chain(*(glob.iglob(os.path.join(folder,'*'+ext)) for ext in (
                    Constants.tobecopied_ext, Constants.copied_ext,
                    Constants.copying_ext, Constants.problematic_ext))):
                if len(file(oldfile).readlines()) < 4: os.remove(oldfile)

        # Handle .COPIED files from previous run
        self.logger.info('Handle .COPIED files from previous run')
        for folder in self.DirList:
            for oldfile in glob.iglob(os.path.join(folder,'*'+Constants.copied_ext)):
                DataFile = os.path.splitext(oldfile)[0]
                CastorInfo = self.getCastorInfo(DataFile)
                CastorPool = CastorInfo[0]
                CastorDir = CastorInfo[1]
                StageHost = CastorInfo[2]

                self.CopyFileList.append(DataFile)
                if self.conf.DeleteEnabled:
                    self.DeleteQueue.put([DataFile,CastorPool,CastorDir,StageHost])
                self.logger.debug('File: %s in DeleteQueue', DataFile)

        self.logger.info('Size of DeleteQueue: %d', self.DeleteQueue.qsize())

        # Get the initial sorted list of files to be copied
        # CopyFileList is a list of (.TOBECOPIED filename, Pool, Castor Directory,StageHost)
        self.logger.info('Get the initial list of files to be copied')

        # Send those files to the copy queue
        for element in self.getCopyFiles():
            # order is important here: first CopyFileList then send to CopyThread
            self.CopyFileList.append(element[0])
            self.CopyQueue.put(element)
            self.logger.debug('File: %s in CopyQueue', element[0])

        self.logger.info('Size of CopyQueue: %d', self.CopyQueue.qsize())


    def loop(self):
        while not self.exitFlag:
            self.event.wait(self.ManagerTimeout)
            self.logger.info('Clear processed files. Size of ClearQueue: %d', self.ClearQueue.qsize())

            clearCounter = 0

            while clearCounter < self.nDel:

                if self.exitFlag: break
                if self.ClearQueue.qsize() == 0: break
                self.logger.debug('clearCounter = %d; nDel = %d', clearCounter, self.nDel)

                # Get file to be cleared from ClearQueue
                clearfile = self.ClearQueue.get(0)

                # Remove clearfile from CopyFileList
                self.CopyFileList.remove(clearfile[0])
                if not glob.glob(clearfile[0] + Constants.problematic_ext):
                    self.ProblDict.pop(clearfile[0], None)

                self.logger.debug('File: %s removed from CopyFileList', clearfile[0])

                # Count the deleted files
                clearCounter += 1

            self.logger.info('Size of ClearQueue: %d', self.ClearQueue.qsize())

            # After deletion of nDel files, update the list of files to be copied and the copy queue
            self.logger.info('Update the list of files to be copied')
            NewList = self.getCopyFiles()
            for element in NewList:
                self.CopyFileList.append(element[0])
                self.CopyQueue.put(element)
                self.logger.debug('File: %s in CopyQueue', element[0])

            self.logger.info('Size of CopyQueue: %d', self.CopyQueue.qsize())

            # Update current year and month for directory streaming
            self.updateDates()


    def run(self):
        self.logger.info(thread_id_string())

        try:
            self.initialize()
            self.loop()

        except IOError, e:
            self.logger.critical('Cannot write Castor Info to %s because: "%s". '
                    'ManagerThread is stopping!', e.filename, e.strerror)

        self.logger.info('ManagerThread exited')


    # Get the new files to be copied
    def getCopyFiles(self):
        date_file_list = []

        for folder in self.DirList:
            # Consider only unlocked filesystems
            if glob.glob(os.path.join(folder,self.lockFile)):
                self.logger.debug('Directory "%s" is locked: do not copy files from there', folder)
                continue

            files = sum([glob.glob(os.path.join(folder,p)) for p in self.DataFilePattern],[])
            self.logger.debug('%d files found on folder "%s" matching "%s"',
                    len(files), folder, pprint.pformat(self.DataFilePattern))

            for filename in files:

                # Protection against bad pattern definition in the previous glob
                if os.path.splitext(filename)[1] in (
                        Constants.tobecopied_ext, Constants.copied_ext,
                        Constants.copying_ext, Constants.problematic_ext):
                    continue

                # Do not take already listed files
                if not filename in self.CopyFileList:
                    self.logger.debug('New file: %s', filename)

                    if self.conf.ExcludeFileRegex is not None and re.match(self.conf.ExcludeFileRegex, filename):
                        self.logger.debug('ignoring file matching exclude regex: %s', filename)
                        continue

                    # Check for minimal file size, if needed
                    if self.MinSizekB:
                        try:
                            filesize = os.path.getsize(filename)
                        except OSError as exc:
                            if exc.errno == errno.ENOENT:
                                # The file has been removed by an external cause.
                                # This is bad but this is not a reason to crash.
                                self.logger.critical('error getting file size for %s: no such file', filename)
                                continue
                            else:
                                raise exc

                        if filesize/1024.< self.MinSizekB:
                            self.logger.debug('File %s is too small. Skipping.', filename)
                            if self.RemoveSmallFiles:
                                self.logger.debug('File %s is too small. Deleting.', filename)
                                try:
                                    os.remove(filename)
                                except OSError, ex:
                                    self.logger.warning('Cannot delete %s. ERROR: %s', filename, ex.__repr__())

                            continue

                    # Check if .PROBLEMATIC file is ready to be copied
                    if glob.glob(filename + Constants.problematic_ext) \
                       and not self.checkProbl(filename):
                        continue

                    # Get Castor pool, copy directory and stage host
                    CastorInfo = self.getCastorInfo(filename)
                    Pool = CastorInfo[0]
                    CopyDir = CastorInfo[1]
                    StageHost = CastorInfo[2]

                    # Get the last modified date (mtime) for file
                    # stat = (mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime)
                    timestamp = os.stat(filename)[8]

                    # Get if file is an Express stream
                    express = self.isExpress(filename)

                    # Create a list of the previous lists
                    date_file_list.append((timestamp, filename, express, Pool, CopyDir, StageHost))

                else:
                    self.logger.debug('File %s already in the internal copy list', filename)

        self.logger.debug('Sort new files, with priority to express streams, then to oldest ones')
        date_file_list.sort(self.Compare)

        return [([liste[1],liste[3],liste[4],liste[5]]) for liste in date_file_list]


    # Find out if a file is an Express stream
    def isExpress(self,filename):
        return 'express' in filename.lower()


    # Compare two (timestamp, filename, is express)
    def Compare(self,a, b):
        if (a[2]!= b[2]): return cmp(-a[2],-b[2])
        else: return cmp(a[0],b[0])


    # ProblDict is a dictionary of {.PROBLEMATIC file : [timestamp,nretry]}
    def checkProbl(self,ProblFile):
        self.logger.debug('Problematic file: %s. Check if it is time to retry the copying', ProblFile)
        if ProblFile not in self.ProblDict:
            self.ProblDict[ProblFile] = [time(),0]
            self.logger.debug('First time: retry')
            return True
        # end if
        timestamp = self.ProblDict.get(ProblFile)[0]
        nretry = self.ProblDict.get(ProblFile)[1]

        #print (time()-timestamp),self.ProblDelay*math.exp(self.ProblScalingFactor*nretry)
        #print len(self.ProblDict)

        # Try to copy the file again only if the timeout is expired
        if (time()-timestamp) > (self.ProblDelay * math.exp(self.ProblScalingFactor*nretry)):
            nretry += 1
            self.ProblDict[ProblFile] = [time(), nretry]
            self.logger.debug('Time to retry (%d times up to now)', nretry)

            # Inform the user if we cannot copy the file and the retry time
            # limit has been reached
            if self.ProblDelay*math.exp(self.ProblScalingFactor*nretry) > self.ProblDelayLimit:
                filesize = os.path.getsize(ProblFile)
                self.logger.critical('Problematic file %s reached the delay retry limit of %d seconds. '
                        'Please have a look! (filesize= %d, copy_timeout= %d)',
                        ProblFile, self.ProblDelayLimit, filesize, self.conf.NoneTimeout)
            return True
        else:
            self.logger.debug('Last retry was too recent: retry later')
            return False


    # Get directory where to copy files in Castor and Castor environment
    def getCastorInfo(self,filename):

        # For files from previous run, read CopyDir and CastorEnv from
        # .TOBECOPIED,.COPYING,.PROBLEMATIC,.COPIED files
        infoFile = ''
        if glob.glob(filename + Constants.tobecopied_ext): infoFile = filename + Constants.tobecopied_ext
        elif glob.glob(filename + Constants.copying_ext): infoFile = filename + Constants.copying_ext
        elif glob.glob(filename + Constants.problematic_ext): infoFile = filename + Constants.problematic_ext
        elif glob.glob(filename + Constants.copied_ext): infoFile = filename + Constants.copied_ext

        if infoFile:
            self.logger.debug('File: %s from previous run. Read Castor Info from %s', filename, infoFile)
            InfoFile = open(infoFile,'r')
            allLines = InfoFile.readlines()
            InfoFile.close()
            CastorDir = [x for x in allLines if 'CastorDir' in x][0].split()[2]
            try:
                pool = [x for x in allLines if 'CastorPool' in x][0].split()[2]
            except IndexError:
                # EOS does not require a pool
                pool = ''
            stagehost = [x for x in allLines if 'StageHost' in x][0].split()[2]
            self.logger.debug('Castor Info: pool = %s, copy directory = %s, stage host = %s', pool, CastorDir, stagehost)

            if not infoFile.endswith(Constants.copied_ext): os.rename(infoFile,filename + Constants.tobecopied_ext)
            return [pool,CastorDir,stagehost]

        try:
            parsed = self.parser(filename)
        except:
            # log to be able to investigate
            self.logger.critical('error parsing filename: %s', filename)
            # backward compatibility: re-raise
            raise

        #Prepare dictionary for directory definition
        dirdict = {'year':self.Year,
                   'month':self.Month,
                   'streamtype':parsed.StreamType(),
                   'streamname':parsed.StreamName(),
                   'runnumber':parsed.RunNr(),
                   'filenr':parsed.FileNr(),
                   'lumiblocknr':parsed.LBNr(),
                   'projecttag':parsed.ProjectTag(),
                   'dataset':parsed.Dataset(),
                   'applicationid':parsed.AppId(),
                   'directory':parsed.Directory(),
                   'userdef1':parsed.UserDef1(),
                   'userdef2':parsed.UserDef2(),
                   'userdef3':parsed.UserDef3()}

        # For new files, read CopyDir and CastorEnv from the configuration and
        # write in .TOBECOPIED file
        self.logger.debug('New file: %s. Read Castor Info from Configuration', filename)

        # Stream driven case
        def match_pool(drivenPool, parsed):
            return (drivenPool.projecttag.match(parsed.ProjectTag()) and
                    drivenPool.streamtype.match(parsed.StreamType()) and
                    drivenPool.streamname.match(parsed.StreamName()) and
                    drivenPool.lumiblock.match(parsed.LBNr()) and
                    drivenPool.application.match(parsed.AppId()) and
                    drivenPool.directory.match(parsed.Directory()))

        if self.drivenPool:
            for e in self.drivenPool:
                if match_pool(e, parsed):
                    self.logger.debug('Stream driven case')
                    CastorDir = os.path.normpath(e.targetdir%dirdict)
                    pool =  e.pool
                    stagehost = e.stagehost
                    self.logger.debug('Castor Info: pool = %s, copy directory = %s, stage host = %s', pool, CastorDir, stagehost)
                    self.writeCastorInfo(filename,pool,CastorDir,stagehost)
                    return [pool,CastorDir,stagehost]

        # Normal case
        self.logger.debug('Normal case')
        pool = self.Pool
        stagehost = self.StageHost
        CastorDir = os.path.normpath(self.CopyDir%dirdict)

        self.logger.debug('Castor Info: pool = %s, copy directory = %s, stage host = %s', pool, CastorDir, stagehost)
        self.writeCastorInfo(filename,pool,CastorDir,stagehost)
        return [pool,CastorDir,stagehost]


    # Write Castor copy directory and environment to .TOBECOPIED file
    def writeCastorInfo(self,filename,Pool,CopyDir,stagehost):
        self.logger.debug('Write Castor Info to %s%s', filename, Constants.tobecopied_ext)
        FileName = 'FileName = ' + filename + '\n'
        CastorPool = 'CastorPool = ' + Pool + '\n'
        CastorDir = 'CastorDir = ' + CopyDir + '\n'
        StageHost = 'StageHost = ' + stagehost + '\n'
        InfoFile = open(filename + Constants.tobecopied_ext, 'w')
        InfoFile.writelines([FileName,CastorPool,CastorDir,StageHost])
        InfoFile.close()


    # Exit handler: exit in clean way
    def managerExit(self):
        self.exitFlag = 1
        self.logger.info('Exit signal received')


    def updateDates(self):
        self.Year = datetime.date.today().year
        self.Month = datetime.date.today().month


    def setFileParser(self, parser):
        self.parser = parser


    # Function to add / remove ERS handler
    def AddRemoveERS(self, create_ers_handler):
        if create_ers_handler:
            self.logger.addHandler( self.manage_ERS_handler )
        else:
            self.logger.removeHandler( self.manage_ERS_handler )
