#!/bin/env python
from cs.Tools.Libraries.config import Config
import importlib
from collections import namedtuple
import re

class REWrap(object):
###
# Wrap python re objects such that a None
# pattern results in a positive match
###
    def __init__(self, pattern):
        if not pattern == None:
            self.re = re.compile(pattern)
        else:
            self.re = None

    def match(self, string):
        if self.re:
            return self.re.match(string)
        else:
            return True


class DrivenPool(namedtuple('DrivenPool', ['targetdir','stagehost','pool','projecttag','streamtype','streamname','lumiblock','application','directory'])):
###
# namedtuple is subclassed to
# a) allow default values (None) for the filtering parameters
# b) crate regex objects from the filtering parameters
###
    def __new__(cls, targetdir, stagehost, pool='',
                projecttag=None, streamtype=None,
                streamname=None, lumiblock=None,
                application=None, directory=None):
        return super(DrivenPool, cls).__new__(cls, targetdir, stagehost,
                                        pool, REWrap(projecttag),
                                        REWrap(streamtype),
                                        REWrap(streamname),
                                        REWrap(lumiblock),
                                        REWrap(application),
                                        REWrap(directory))

class Conf:

    def __init__(self, cfgfile):

        cfg = Config(file(cfgfile))

        ########## Common parameters ##########

        # Log directory
        self.LogDir = cfg.LogDir

        #Lock file
        self.lockFile = cfg.lockFile

        # Log severity level (DEBUG,INFO,WARNING,ERROR,CRITICAL)
        self.LogLevel = cfg.LogLevel

        # ERS severity level (DEBUG,INFO,WARNING,ERROR,CRITICAL)
        try:
            self.ERSLogLevel = cfg.ERSLogLevel
        except AttributeError:
            self.ERSLogLevel = 'info'

        # On / Off switch for ERS
        try:
            self.ERSenabled = cfg.ERSenabled
        except AttributeError:
            self.ERSenabled = False

        # Email list which will receive error messages (list of strings)
        self.mailList = cfg.mailList

        # Email Log severity level (DEBUG,INFO,WARNING,ERROR,CRITICAL)
        self.mailLevel = cfg.mailLevel

        # Email sender
        self.mailSender = cfg.mailSender

        # PYTHON_PATH
        self.pythonPathPrepend = []
        try:
            self.pythonPathPrepend = cfg.pythonPathPrepend
        except AttributeError:
            pass


        ########## Variables for ERS ##########

        # Partition to log to
        try:
            self.partition = cfg.partition
        except AttributeError:
            self.partition = ''

        # Stream ers to
        try:
            self.ers_error = cfg.ers_error
        except AttributeError:
            self.ers_error = ''

        try:
            self.ers_info = cfg.ers_info
        except AttributeError:
            self.ers_info = ''

        try:
            self.ers_warning = cfg.ers_warning
        except AttributeError:
            self.ers_warning = ''

        try:
            self.ers_debug = cfg.ers_debug
        except AttributeError:
            self.ers_debug = ''


        ########## METADATA DATABASE ##########

        # Oracle connection string
        # Use None to disable DB access
        # To avoid having credential in plain text in config files, we now rely
        # on authentication.xml file containing credentials.
        # The authentication.xml is looked up in $CORAL_AUTH_PATH.
        #
        # Example:
        #   self.connection = 'oracle://atonr_conf/ATLAS_SFO_T0'
        self.connection = cfg.connection

        # File table name
        self.file_table = cfg.file_table

        # Lumiblock table name
        self.lb_table = cfg.lb_table

        # Run table name
        self.run_table = cfg.run_table


        ########## MAIN THREAD ##########

        # If Oracle database is down, retry to connect to it every
        # DBTimeout (s)
        self.DBTimeout = cfg.DBTimeout

        # Timeout to refresh the Oracle connection
        self.DBReconnectTimeout = cfg.DBReconnectTimeout

        # Filename parser module
        self.Filenameparser = cfg.Filenameparser

        # Filename parser configuration
        try:
            self.FilenameparserConf = dict(cfg.FilenameparserConf.iteritems())
        except AttributeError:
            self.FilenameparserConf = {}

        # Backend module
        try:
            self.backend = importlib.import_module("cs.StorageBackends." + cfg.BackendModule)
        except AttributeError:
            #Just for configuration backcompatibility
            self.backend = importlib.import_module('cs.StorageBackends.castorstorage')

        #Backend params
        try:
            self.backend.init(dict(cfg.BackendModuleConf.iteritems()))
        except AttributeError:
            pass

        #Kerberos Auth
        try:
            self.keytab = cfg.Keytab
        except AttributeError:
            self.keytab = None

        try:
            self.krbuser = cfg.KrbUser
        except AttributeError:
            self.krbuser = None

        try:
            self.krbcache = cfg.KrbCache
        except AttributeError:
            self.krbcache = None


        ########## MANAGER THREAD ##########

        # Directory List where to get Data Files (list of strings)
        self.DirList = cfg.DirList

        # Unix pathname pattern for file selection
        # self.FilePattern = '*[.data,.out]'
        self.DataFilePattern = cfg.DataFilePattern

        try:
            self.ExcludeFileRegex = cfg.ExcludeFileRegex
        except AttributeError:
            self.ExcludeFileRegex = None

        # Number of files to be deleted before to update the list of files
        # to be copied
        self.nDel = cfg.nDel

        # Sleep timeout (s) before updating the list of files to be copied
        self.ManagerTimeout = cfg.ManagerTimeout

        # Delay (s) before to put .PROBLEMATIC file back into the CopyQueue
        self.ProblDelay = cfg.ProblDelay

        # At each retry, the delay for .PROBLEMATIC files is scaled
        # using this factor:
        # ActualProblDelay = ProblDelay*e^(ProblScalingFactor*#retries)
        self.ProblScalingFactor = cfg.ProblScalingFactor

        # Actual .PROBLEMATIC delay limit to enable CRITICAL messages
        self.ProblDelayLimit = cfg.ProblDelayLimit

        # Minimal file size (in kB)
        try:
            self.MinSizekB = cfg.MinSizekB
        except AttributeError:
            self.MinSizekB = None

        #If small files should be removed
        try:
            self.RemoveSmallFiles = cfg.RemoveSmallFiles
        except AttributeError:
            self.RemoveSmallFiles = False

        # Directory where to copy the Data Files (string)
        #Keywords can be used here
        self.CopyDir = cfg.CopyDir

        # Pool in Castor
        self.Pool = cfg.Pool

        # Stage Host
        self.StageHost = cfg.StageHost


        # Stream driven pool
        # [{'targetdir':'/some/where', 'stagehost':'castoratlas',
        # 'pool':'t0atlas', 'projecttag':'data_test', 'streamtype':'physics',
        # 'streamname':'something' ,'lumiblock':'22?', 'application':'SFO*',
        # 'directory':'/some/dir'},]
        self.drivenPool = []
        for i in cfg.drivenPool:
            try:
                d = dict(i.iteritems())
            except AttributeError,e:
                # Perhaps it is an old configuration file style
                # [(projecttag, stream type, stream name, pool,
                # copy directory, stagehost),...]
                tags= ['projecttag','streamtype','streamname','pool',
                       'targetdir','stagehost']
                if len(tags) != len(i):
                    raise e
                d = dict((t,e) for t,e in zip(tags,i) if e)

            try:
                self.drivenPool.append(DrivenPool(**d))
            except TypeError,e:
                print 'Parsing driven pool entry failed: %s' % str(i)
                raise e

        ########## COPY THREAD ##########

        # Max number of concurrent cp processes
        self.maxCopy = cfg.maxCopy

        # In case of failure, max number of rfcp retries on the spot
        # (before let the Manager take care of retring)
        self.maxRetry = cfg.maxRetry

        # Timeout (s) after which kill and retry an rfcp command in None status
        self.NoneTimeout = cfg.NoneTimeout

        # Sleep timeout (s) before checking again the copy process status
        self.CopyTimeout = cfg.CopyTimeout

        # Kill copy processes if FS gets locked
        try:
            self.avoidCopyOverlap = cfg.avoidCopyOverlap
        except AttributeError:
            self.avoidCopyOverlap = False

        # Compute adler32 checksum from the file
        try:
            self.computeChecksum = cfg.computeChecksum
        except AttributeError:
            self.computeChecksum = False

        # Bandwidth limit
        try:
            self.bwLimit = cfg.bwLimit
        except AttributeError:
            self.bwLimit = 0

        # Bandwidth device
        try:
            self.bwDevice = cfg.bwDevice
        except AttributeError:
            self.bwDevice = None


        ########## DELETE THREAD ##########
        try:
            self.DeleteEnabled = cfg.DeleteEnabled
        except AttributeError:
            self.DeleteEnabled = True

        # Is simultaneous writing and deleting allowed?
        self.ignoreLock = cfg.ignoreLock

        # Get filesystem info after nFiles have been copied
        self.nFiles = cfg.nFiles

        # Do not delete files more recent than minFileTime (s)
        self.minFileTime = cfg.minFileTime

        # Is migration required before deleting?
        self.migFlag = cfg.migFlag

        # Module that will provide the merged file name builder function
        self.MergedNameModule = cfg.MergedNameModule
        if self.MergedNameModule:
            self.MergedModule = importlib.import_module("cs.Tools.MigrationCheckers." + cfg.MergedNameModule)
            # Merge module params
            try:
                self.MergedModule.init(dict(cfg.MergedModuleConf.iteritems()))
            except AttributeError:
                pass
        else:
            self.MergedModule = None

        # Perform checksum check
        self.deletionChecksumCheck = cfg.checksumCheck

        # Max total filesystem usage allowed while waiting for migration (in %)
        self.highCriticMark = cfg.highCriticMark

        # Min total filesystem usage allowed while waiting for migration (in %)
        self.lowCriticMark = cfg.lowCriticMark

        # Low watermark: do not delete if filesystems
        # usage in DirList below threshold (in %)
        self.lowWaterMark = [i for i in cfg.lowWaterMark]

        # Sleep timeout (s) before deleting next file in DeleteQueue
        self.DeleteTimeout = cfg.DeleteTimeout

        # Delay before to check if file is migrated to Castor tape (s)
        self.MigDelay = cfg.MigDelay


        # Crosschecks
        self.lowWaterMark.extend([0]* \
                                 (len(self.DirList)-len(self.lowWaterMark)))

        ########## CHECK THREAD ##########
        try:
            self.CheckTimeout = cfg.CheckTimeout
        except AttributeError:
            self.CheckTimeout = 1

        try:
            self.ERSTimeout = cfg.ERSTimeout
        except AttributeError:
            self.ERSTimeout = 10

        ## DDM MONITORING
        try:
            self.DdmMonitoringEnabled = cfg.DdmMonitoringEnabled
        except AttributeError:
            self.DdmMonitoringEnabled = False

        try:
            self.DdmMonitoringTimeout = cfg.DdmMonitoringTimeout
        except AttributeError:
            self.DdmMonitoringTimeout = 2

        try:
            self.DdmMonitoringProxy = cfg.DdmMonitoringProxy
        except AttributeError:
            self.DdmMonitoringProxy = 'atlasgw-exp:3128'

        try:
            self.DdmMonitoringEndpoint = cfg.DdmMonitoringEndpoint
        except AttributeError:
            self.DdmMonitoringEndpoint = 'https://rucio-lb-prod.cern.ch/traces/'

        try:
            self.DdmMonitoringQueueMaxSize = cfg.DdmMonitoringQueueMaxSize
        except AttributeError:
            self.DdmMonitoringQueueMaxSize = 200

        try:
            self.DBLookupMaxRetries = self.DBLookupMaxRetries
        except AttributeError:
            self.DBLookupMaxRetries = 5
